"""Data models for generating visualization in Kubeflow pipelines UI."""
from typing import List, Union, Optional

from pydantic import Field, BaseModel

from kfx.vis.enums import KfpStorage, KfpVisType, KfpDataType, KfpArtifactDataFormat


class KfpArtifactSchema(BaseModel):
    """Schema for columnar data inside a kubeflow pipeline artifact."""

    name: str
    type: KfpDataType


class KfpVis(BaseModel):
    """Parameters to generate a visualization inside kubeflow pipeline UI."""

    artifact_format: Optional[KfpArtifactDataFormat] = Field(
        None,
        alias="format",
        description="The format of the artifact data. The default is `csv`. "
        "Note: The only format currently available is `csv`.",
    )
    header: Optional[List[str]] = Field(
        None,
        description="A list of strings to be used as headers for the artifact data. "
        "For example, in a table these strings are used in the first row.",
    )
    labels: Optional[List[str]] = Field(
        None,
        description="A list of strings to be used as labels for artifact columns or "
        "rows.",
    )
    predicted_col: Optional[str] = Field(
        None, description="Name of the predicted column."
    )
    artifact_schema: Optional[List[KfpArtifactSchema]] = Field(  # type: ignore
        None,
        alias="schema",
        description="A list of `{type, name}` objects that specify the schema of the "
        "artifact data.",
    )
    source: Optional[str] = Field(
        None,
        description="The full path to the data. The available locations include "
        "`http`, `https`, `Amazon S3`, `Minio`, and `Google Cloud Storage`. "
        "The path can contain wildcards `*`, in which case the Kubeflow Pipelines "
        "UI concatenates the data from the matching source files. "
        "For some viewers, this field can contain inlined string data instead of a "
        "path.",
    )
    storage: Optional[KfpStorage] = Field(
        None,
        description="Applies only to outputs of type `markdown`. "
        "Set to `inline` if `source` contains the actual markdown instead of a "
        "remote source.",
    )
    target_col: Optional[str] = Field(None, description="Name of the target column.")
    type: KfpVisType = Field(
        None,
        description="Types of visualization. "
        "e.g. confustion_matrix, markdown, roc, table, tensorboard, or web-app.",
    )

    class Config:
        """Config model to use alias as schema is a protected key for pydantic."""

        allow_population_by_field_name = True


class ConfusionMatrix(KfpVis):
    """Visualizes a confusion matrix.

    The confusion_matrix viewer plots a confusion matrix visualization
    of the data from the given source path, using the schema to parse
    the data. The labels provide the names of the classes to be plotted
    on the x and y axes.
    """

    type: KfpVisType = Field(
        KfpVisType.confusion_matrix,
        const=True,
        description="Types of visualization. "
        "e.g. confustion_matrix, markdown, roc, table, tensorboard, or web-app.",
    )
    source: str = Field(
        ...,
        description="The full path to the data. The available locations include "
        "`http`, `https`, `Amazon S3`, `Minio`, and `Google Cloud Storage`. "
        "The path can contain wildcards `*`, in which case the Kubeflow Pipelines "
        "UI concatenates the data from the matching source files. "
        "For some viewers, this field can contain inlined string data instead of a "
        "path.",
    )
    artifact_format: KfpArtifactDataFormat = Field(
        KfpArtifactDataFormat.csv,
        alias="format",
        description="The format of the artifact data. The default is `csv`. "
        "Note: The only format currently available is `csv`.",
    )
    artifact_schema: List[KfpArtifactSchema] = Field(  # type: ignore
        [
            {"name": "target", "type": "CATEGORY"},
            {"name": "predicted", "type": "CATEGORY"},
            {"name": "count", "type": "NUMBER"},
        ],
        alias="schema",
        description="A list of `{type, name}` objects that specify the schema of the "
        "artifact data.",
    )
    labels: List[str] = Field(
        ...,
        description="A list of strings to be used as labels for artifact columns or "
        "rows.",
    )


class Markdown(KfpVis):
    """Visualizes a markdown.

    The markdown viewer renders Markdown strings on the Kubeflow Pipelines UI.
    The viewer can read the Markdown data from the following locations:

    -   A Markdown-formatted string embedded in the source field. The value of the
        storage field must be `inline`.

    -   Markdown code in a remote file, at a path specified in the source field.
        The storage field can contain any value except `inline`.
    """

    type: KfpVisType = Field(
        KfpVisType.markdown,
        const=True,
        description="Types of visualization. "
        "e.g. confustion_matrix, markdown, roc, table, tensorboard, or web-app.",
    )
    source: str = Field(
        ...,
        description="The full path to the data. The available locations include "
        "`http`, `https`, `Amazon S3`, `Minio`, and `Google Cloud Storage`. "
        "The path can contain wildcards `*`, in which case the Kubeflow Pipelines "
        "UI concatenates the data from the matching source files. "
        "For some viewers, this field can contain inlined string data instead of a "
        "path.",
    )
    storage: Optional[KfpStorage] = Field(
        None,
        description="Applies only to outputs of type `markdown`. "
        "Set to `inline` if `source` contains the actual markdown instead of a "
        "remote source.",
    )


class Roc(KfpVis):
    """Visualizes a ROC curve.

    The roc viewer plots a receiver operating characteristic (ROC) curve using the
    data from the given source path. The Kubeflow Pipelines UI assumes that the schema
    includes three columns with the following names:

    - fpr (false positive rate)
    - tpr (true positive rate)
    - thresholds

    When viewing the ROC curve, you can hover your cursor over the ROC curve to see
    the threshold value used for the cursor’s closest fpr and tpr values.
    """

    type: KfpVisType = Field(
        KfpVisType.roc,
        const=True,
        description="Types of visualization. "
        "e.g. confustion_matrix, markdown, roc, table, tensorboard, or web-app.",
    )
    artifact_format: KfpArtifactDataFormat = Field(
        KfpArtifactDataFormat.csv,
        alias="format",
        description="The format of the artifact data. The default is `csv`. "
        "Note: The only format currently available is `csv`.",
    )
    artifact_schema: List[KfpArtifactSchema] = Field(  # type: ignore
        [
            {"name": "fpr", "type": "NUMBER"},
            {"name": "tpr", "type": "NUMBER"},
            {"name": "thresholds", "type": "NUMBER"},
        ],
        alias="schema",
        description="A list of `{type, name}` objects that specify the schema of the "
        "artifact data.",
    )
    source: str = Field(
        ...,
        description="The full path to the data. The available locations include "
        "`http`, `https`, `Amazon S3`, `Minio`, and `Google Cloud Storage`. "
        "The path can contain wildcards `*`, in which case the Kubeflow Pipelines "
        "UI concatenates the data from the matching source files. "
        "For some viewers, this field can contain inlined string data instead of a "
        "path.",
    )


class Table(KfpVis):
    """Visualizes a table.

    The table viewer builds an HTML table out of the data at the given source path,
    where the header field specifies the values to be shown in the first row of the
    table.

    The table supports pagination.
    """

    type: KfpVisType = Field(
        KfpVisType.table,
        const=True,
        description="Types of visualization. "
        "e.g. confustion_matrix, markdown, roc, table, tensorboard, or web-app.",
    )
    artifact_format: KfpArtifactDataFormat = Field(
        KfpArtifactDataFormat.csv,
        alias="format",
        description="The format of the artifact data. The default is `csv`. "
        "Note: The only format currently available is `csv`.",
    )
    header: List[str] = Field(
        ...,
        description="A list of strings to be used as headers for the artifact data. "
        "For example, in a table these strings are used in the first row.",
    )
    source: str = Field(
        ...,
        description="The full path to the data. The available locations include "
        "`http`, `https`, `Amazon S3`, `Minio`, and `Google Cloud Storage`. "
        "The path can contain wildcards `*`, in which case the Kubeflow Pipelines "
        "UI concatenates the data from the matching source files. "
        "For some viewers, this field can contain inlined string data instead of a "
        "path.",
    )


class Tensorboard(KfpVis):
    """Provides a tensorboard viewer.

    The tensorboard viewer adds a Start Tensorboard button to the output page.

    When viewing the output page, you can:

    -   Click Start Tensorboard to start a TensorBoard Pod in your Kubeflow cluster.
        The button text switches to Open Tensorboard.

    -   Click Open Tensorboard to open the TensorBoard interface in a new tab,
        pointing to the logdir data specified in the source field.

    Note:
    The Kubeflow Pipelines UI doesn’t fully manage your TensorBoard instances.
    The “Start Tensorboard” button is a convenience feature so that you don’t have
    to interrupt your workflow when looking at pipeline runs. You’re responsible
    for recycling or deleting the TensorBoard Pods using your Kubernetes management
    tools.
    """

    type: KfpVisType = Field(
        KfpVisType.tensorboard,
        const=True,
        description="Types of visualization. "
        "e.g. confustion_matrix, markdown, roc, table, tensorboard, or web-app.",
    )
    source: str = Field(
        ...,
        description="The full path to the data. The available locations include "
        "`http`, `https`, `Amazon S3`, `Minio`, and `Google Cloud Storage`. "
        "The path can contain wildcards `*`, in which case the Kubeflow Pipelines "
        "UI concatenates the data from the matching source files. "
        "For some viewers, this field can contain inlined string data instead of a "
        "path.",
    )


class WebApp(KfpVis):
    """Provides a web-app viewer.

    The web-app viewer provides flexibility for rendering custom output.
    You can specify an HTML file that your component creates, and the
    Kubeflow Pipelines UI renders that HTML in the output page. The HTML
    file must be self-contained, with no references to other files in
    the filesystem. The HTML file can contain absolute references to
    files on the web. Content running inside the web app is isolated in
    an iframe and cannot communicate with the Kubeflow Pipelines UI.
    """

    type: KfpVisType = Field(
        KfpVisType.web_app,
        const=True,
        description="Types of visualization. "
        "e.g. confustion_matrix, markdown, roc, table, tensorboard, or web-app.",
    )
    source: str = Field(
        ...,
        description="The full path to the data. The available locations include "
        "`http`, `https`, `Amazon S3`, `Minio`, and `Google Cloud Storage`. "
        "The path can contain wildcards `*`, in which case the Kubeflow Pipelines "
        "UI concatenates the data from the matching source files. "
        "For some viewers, this field can contain inlined string data instead of a "
        "path.",
    )


class KfpUiMetadata(BaseModel):
    """Describes the visualization to render inside kubeflow pipeline UI."""

    version: Union[int, str] = Field(
        1, description="Version of the kubeflow pipeline ui metadata schema."
    )
    outputs: List[
        Union[ConfusionMatrix, Roc, Markdown, Table, Tensorboard, WebApp]
    ] = Field(
        [], description="List of objects describing the desired kfp visualizations."
    )



"""Helper functions for generating visualization in Kubeflow pipelines UI."""
from typing import List, Union, Optional

from pydantic import BaseModel

from kfx.vis.models import (
    Roc,
    Table,
    WebApp,
    Markdown,
    KfpStorage,
    Tensorboard,
    KfpUiMetadata,
    ConfusionMatrix,
    KfpArtifactDataFormat,
)

KFP_UI_METADATA_PATH = "/mlpipeline-ui-metadata.json"


def confusion_matrix(
    source: str,
    labels: List[str],
    artifact_format: Union[KfpArtifactDataFormat, str] = "csv",
    **kwargs,
) -> ConfusionMatrix:
    """Helper function to create a KfpUiMetadata ConfusionMatrix object.

    The source artifact must be a CSV with the following 3 columns:
    - target
    - predicted
    - count

    Args:
        source (str): Full path to the data artifact.
        labels (List[str]): Names of the classes to be plotted on the x and y axes.
        artifact_format (Union[KfpArtifactDataFormat, str], optional):
            Data format for the artifact. Defaults to "csv".

    Returns:
        ConfusionMatrix: pydantic data object.
    """
    return ConfusionMatrix(
        source=source, artifact_format=artifact_format, labels=labels, **kwargs
    )


def markdown(
    source: str, storage: Optional[Union[KfpStorage, str]] = None, **kwargs
) -> Markdown:
    """Helper function to create a KfpUiMetadata Markdown object.

    Args:
        source (str): Full path to the markdown or the actual markdown
            text.
        storage (Optional[Union[KfpStorage, str]], optional): Set "inline"
            if source has the actual markdown text. Defaults to None.

    Returns:
        Markdown: pydantic data object.
    """
    return Markdown(source=source, storage=storage, **kwargs)


def roc(
    source: str, artifact_format: Union[KfpArtifactDataFormat, str] = "csv", **kwargs
) -> Roc:
    """Helper function to create a KfpUiMetadata Roc object.

    The source artifact must be a CSV with the following 3 columns:
    - fpr (false positive rate)
    - tpr (true positive rate)
    - thresholds

    Args:
        source (str): Full path to roc data.
        artifact_format (Union[KfpArtifactDataFormat, str], optional):
            Data format for the artifact. Defaults to "csv".

    Returns:
        Roc: pydantic data object.
    """
    return Roc(source=source, artifact_format=artifact_format, **kwargs)


def table(
    source: str,
    header: List[str],
    artifact_format: Union[KfpArtifactDataFormat, str] = "csv",
    **kwargs,
) -> Table:
    """Helper function to create a KfpUiMetadata Table object.

    Args:
        source (str): Full path to the data.
        header (List[str]): Headers to use for the table.
        artifact_format (Union[KfpArtifactDataFormat, str], optional):
            Data format for the artifact. Defaults to "csv".

    Returns:
        Table: pydantic data object.
    """
    return Table(
        source=source, header=header, artifact_format=artifact_format, **kwargs
    )


def tensorboard(source: str, **kwargs) -> Tensorboard:
    """Helper function to create a KfpUiMetadata Tensorboard object.

    Args:
        source (str): The full path to the tensorboard logs. Supports * wildcards.

    Returns:
        Tensorboard: pydantic data object.
    """
    return Tensorboard(source=source, **kwargs)


def web_app(source: str, **kwargs) -> WebApp:
    """Helper function to create a KfpUiMetadata WebApp object.

    Args:
        source (str): The full path to the html content or inlined html.

    Returns:
        WebApp: pydantic data object.
    """
    return WebApp(source=source, **kwargs)


def kfp_ui_metadata(
    outputs: List[
        Union[ConfusionMatrix, Roc, Markdown, Table, Tensorboard, WebApp, dict]
    ],
    version: Union[int, str] = 1,
) -> KfpUiMetadata:
    """Helper function to create a KfpUiMetadata object.

    Args:
        outputs (List[Union[ConfusionMatrix, Roc, Markdown, Table, Tensorboard,
            WebApp, dict]]): List of KfpVis objects.
        version (Union[int, str], optional): Schema version. Defaults to 1.

    Returns:
        KfpUiMetadata: pydantic data object.
    """
    return KfpUiMetadata(version=version, outputs=outputs)


def asdict(obj: BaseModel) -> dict:
    """Returns the dict representations of the pydantic data object."""
    return obj.dict(exclude_none=True, by_alias=True)


def asjson(obj: BaseModel) -> str:
    """Return the JSON string representation of the pydantic data object."""
    return obj.json(exclude_none=True, by_alias=True)


def tolocalfile(obj: BaseModel, dst: str = KFP_UI_METADATA_PATH):
    """Writes a pydantic data object as a json file in local file system.

    Args:
        obj (BaseModel): pydantic data object.
        dst (str, optional): Destination path. Defaults to
            "/mlpipeline-ui-metadata.json".
    """
    with open(dst, "w") as fileout:
        fileout.write(asjson(obj))



"""Enums used for kfx.vis."""
from enum import Enum


class KfpVisType(str, Enum):
    """Types of visualization available inside kubeflow pipeline UI."""

    confusion_matrix = "confusion_matrix"
    markdown = "markdown"
    roc = "roc"
    table = "table"
    tensorboard = "tensorboard"
    web_app = "web-app"


class KfpArtifactDataFormat(str, Enum):
    """Supported data format for kubeflow pipeline data artifact for visualization."""

    csv = "csv"


class KfpDataType(str, Enum):
    """Supported data type for data column inside a kubeflow pipeline artifact."""

    CATEGORY = "CATEGORY"
    NUMBER = "NUMBER"
    KEY = "KEY"
    TEXT = "TEXT"
    IMAGE_URL = "IMAGE_URL"


class KfpStorage(str, Enum):
    """Storage medium for visualization source."""

    inline = "inline"
    gcs = "gcs"
    minio = "minio"
    s3 = "s3"
    http = "http"
    https = "https"



"""Tests for kfx.lib.vis."""
import kfx.vis._helpers as kfxvis


def test_confusion_matrix():
    expected = {
        "type": "confusion_matrix",
        "format": "csv",
        "schema": [
            {"name": "target", "type": "CATEGORY"},
            {"name": "predicted", "type": "CATEGORY"},
            {"name": "count", "type": "NUMBER"},
        ],
        "source": "gs://your_project/your_bucket/your_cm_file",
        "labels": ["True", "False"],
    }
    data = kfxvis.confusion_matrix(
        source="gs://your_project/your_bucket/your_cm_file", labels=["True", "False"]
    )
    assert kfxvis.asdict(data) == expected, "generates json for confusion matrix"


def test_markdown_inline():
    expected = {
        "storage": "inline",
        "source": "# Inline Markdown\n[A link](https://www.kubeflow.org/)",
        "type": "markdown",
    }
    data = kfxvis.markdown(
        "# Inline Markdown\n[A link](https://www.kubeflow.org/)", storage="inline"
    )
    assert kfxvis.asdict(data) == expected, "generates json for inline markdown"


def test_markdown_file():
    expected = {
        "source": "gs://your_project/your_bucket/your_markdown_file",
        "type": "markdown",
    }
    data = kfxvis.markdown("gs://your_project/your_bucket/your_markdown_file")
    assert kfxvis.asdict(data) == expected, "generates json for file-based markdown"


def test_roc():
    expected = {
        "type": "roc",
        "format": "csv",
        "schema": [
            {"name": "fpr", "type": "NUMBER"},
            {"name": "tpr", "type": "NUMBER"},
            {"name": "thresholds", "type": "NUMBER"},
        ],
        "source": "gs://your_project/your_bucket/your_roc_file",
    }
    data = kfxvis.roc("gs://your_project/your_bucket/your_roc_file")
    assert kfxvis.asdict(data) == expected, "generates json for roc"


def test_table():
    expected = {
        "type": "table",
        "storage": "gcs",
        "format": "csv",
        "header": ["col1", "col2"],
        "source": "gs://your_project/your_bucket/your_csv_file",
    }
    data = kfxvis.table(
        "gs://your_project/your_bucket/your_csv_file",
        header=["col1", "col2"],
        storage="gcs",
    )
    assert kfxvis.asdict(data) == expected, "generates json for table"


def test_tensorboard():
    expected = {"type": "tensorboard", "source": "gs://your_project/your_bucket/logs/*"}
    data = kfxvis.tensorboard("gs://your_project/your_bucket/logs/*")
    assert kfxvis.asdict(data) == expected, "generates json for tensorboard"


def test_webapp():
    expected = {
        "type": "web-app",
        "source": "gs://your_project/your_bucket/your_html_file",
        "storage": "gcs",
    }
    data = kfxvis.web_app("gs://your_project/your_bucket/your_html_file", storage="gcs")
    assert kfxvis.asdict(data) == expected, "generates json for web app"


def test_ui_metadata():
    expected = {
        "version": 1,
        "outputs": [
            {
                "type": "confusion_matrix",
                "format": "csv",
                "schema": [
                    {"name": "target", "type": "CATEGORY"},
                    {"name": "predicted", "type": "CATEGORY"},
                    {"name": "count", "type": "NUMBER"},
                ],
                "source": "gs://your_project/your_bucket/your_cm_file",
                "labels": ["True", "False"],
            },
            {
                "storage": "inline",
                "source": "# Inline Markdown\n[A link](https://www.kubeflow.org/)",
                "type": "markdown",
            },
            {
                "source": "gs://your_project/your_bucket/your_markdown_file",
                "type": "markdown",
            },
            {
                "type": "roc",
                "format": "csv",
                "schema": [
                    {"name": "fpr", "type": "NUMBER"},
                    {"name": "tpr", "type": "NUMBER"},
                    {"name": "thresholds", "type": "NUMBER"},
                ],
                "source": "gs://your_project/your_bucket/your_roc_file",
            },
            {
                "type": "table",
                "storage": "gcs",
                "format": "csv",
                "header": ["col1", "col2"],
                "source": "gs://your_project/your_bucket/your_csv_file",
            },
            {"type": "tensorboard", "source": "gs://your_project/your_bucket/logs/*"},
            {
                "type": "web-app",
                "source": "gs://your_project/your_bucket/your_html_file",
                "storage": "gcs",
            },
        ],
    }
    data = kfxvis.kfp_ui_metadata(
        [
            kfxvis.confusion_matrix(
                source="gs://your_project/your_bucket/your_cm_file",
                labels=["True", "False"],
            ),
            kfxvis.markdown(
                "# Inline Markdown\n[A link](https://www.kubeflow.org/)",
                storage="inline",
            ),
            kfxvis.markdown("gs://your_project/your_bucket/your_markdown_file"),
            kfxvis.roc("gs://your_project/your_bucket/your_roc_file"),
            kfxvis.table(
                "gs://your_project/your_bucket/your_csv_file",
                header=["col1", "col2"],
                storage="gcs",
            ),
            kfxvis.tensorboard("gs://your_project/your_bucket/logs/*"),
            kfxvis.web_app(
                "gs://your_project/your_bucket/your_html_file", storage="gcs"
            ),
        ]
    )
    assert kfxvis.asdict(data) == expected, "generates json for kfp ui metadata"



"""Module for compatibilities with potential updates in dependent packages."""
import re

from kfp.compiler._k8s_helper import sanitize_k8s_name as _sanitize_k8s_name


def __sanitize_k8s_name(name, allow_capital_underscore=False):
    """sanitize_k8s_name cleans and converts the names in the workflow.

    NOTE
    This is copied from the main kfp package which is from _make_kubernetes_name

    Args:
      name: original name,
      allow_capital_underscore: whether to allow capital letter and underscore
        in this name.

    Returns:
      sanitized name.
    """
    if allow_capital_underscore:
        return (
            re.sub("-+", "-", re.sub("[^-_0-9A-Za-z]+", "-", name))
            .lstrip("-")
            .rstrip("-")
        )
    return (
        re.sub("-+", "-", re.sub("[^-0-9a-z]+", "-", name.lower()))
        .lstrip("-")
        .rstrip("-")
    )


sanitize_k8s_name = (  # pylint: disable=invalid-name
    _sanitize_k8s_name or __sanitize_k8s_name
)



"""Extensions for visualizations in kubeflow.

::

    import kfp.components
    import kfx.vis

    from kfx.vis.enums import KfpStorage


    @func_to_container_op
    def some_op(mlpipeline_ui_metadata: OutputTextFile(str)):
        "kfp operator that provides metadata for visualizations."

        mlpipeline_ui_metadata = kfx.vis.kfp_ui_metadata(
            [
                # creates a confusion matrix vis
                kfx.vis.confusion_matrix(
                    source="gs://your_project/your_bucket/your_cm_file",
                    labels=["True", "False"],
                ),
                # creates a markdown with inline source
                kfx.vis.markdown(
                    "# Inline Markdown: [A link](https://www.kubeflow.org/)",
                    storage="inline",
                ),
                # creates a markdown with a remote source
                kfx.vis.markdown(
                    "gs://your_project/your_bucket/your_markdown_file",
                ),
                # creates a ROC curve with a remote source
                kfx.vis.roc(
                    "gs://your_project/your_bucket/your_roc_file",
                ),
                # creates a Table with a remote source
                kfx.vis.table(
                    "gs://your_project/your_bucket/your_csv_file",
                    header=["col1", "col2"],
                ),
                # creates a tensorboard viewer
                kfx.vis.tensorboard(
                    "gs://your_project/your_bucket/logs/*",
                ),
                # creates a custom web app from a remote html file
                kfx.vis.web_app(
                    "gs://your_project/your_bucket/your_html_file",
                ),
            ]
        )

        # write ui metadata so that kubeflow pipelines UI can render visualizations
        mlpipeline_ui_metadata.write(kfx.vis.asjson(mlpipeline_ui_metadata))

"""
from kfx.vis._helpers import (
    roc,
    table,
    asdict,
    asjson,
    web_app,
    markdown,
    tensorboard,
    tolocalfile,
    kfp_ui_metadata,
    confusion_matrix,
)



"""Tests for kfx.lib.utils."""
import os

from typing import List

import kfp.dsl
import kfp.compiler
import kfp.components
import kubernetes.client as k8s_client

from kfp.compiler import Compiler
from kfp.components import OutputTextFile

import kfx.dsl._artifact_location


def is_envs_similar(
    envs: List[k8s_client.V1EnvVar], expected_envs: List[k8s_client.V1EnvVar]
) -> bool:

    return [env.to_dict() for env in envs] == expected_envs


def test_set_workflow_name(tmp_path):
    expected = [
        {"name": "WORKFLOW_NAME", "value": "{{workflow.name}}", "value_from": None}
    ]

    @kfp.components.func_to_container_op
    def test_op():
        import os

        print(os.environ.get("WORKFLOW_NAME"))

    @kfp.dsl.pipeline()
    def test_pipeline():
        op: kfp.dsl.ContainerOp = test_op()
        op.apply(
            kfx.dsl._artifact_location.set_workflow_env(
                kfx.dsl._artifact_location.WorkflowVars(
                    "WORKFLOW_NAME", template="{{workflow.name}}"
                )
            )
        )

        assert is_envs_similar(op.container.env, expected)

    outfile = tmp_path / "pipeline.yaml"
    Compiler().compile(test_pipeline, str(outfile))


def test_set_pod_metadata(tmp_path):
    expected = [
        {
            "name": "POD_NAME",
            "value": None,
            "value_from": {
                "config_map_key_ref": None,
                "field_ref": {"api_version": None, "field_path": "metadata.name"},
                "resource_field_ref": None,
                "secret_key_ref": None,
            },
        },
        {
            "name": "NAMESPACE",
            "value": None,
            "value_from": {
                "config_map_key_ref": None,
                "field_ref": {"api_version": None, "field_path": "metadata.namespace"},
                "resource_field_ref": None,
                "secret_key_ref": None,
            },
        },
        {
            "name": "NODE_NAME",
            "value": None,
            "value_from": {
                "config_map_key_ref": None,
                "field_ref": {"api_version": None, "field_path": "spec.nodeName"},
                "resource_field_ref": None,
                "secret_key_ref": None,
            },
        },
    ]

    @kfp.components.func_to_container_op
    def test_op():
        import os

        print(os.environ.get("WORKFLOW_NAME"))

    @kfp.dsl.pipeline()
    def test_pipeline():
        op: kfp.dsl.ContainerOp = test_op()
        op.apply(kfx.dsl._artifact_location.set_pod_metadata_envs())

        assert is_envs_similar(op.container.env, expected)

    outfile = tmp_path / "pipeline.yaml"
    Compiler().compile(test_pipeline, str(outfile))


def test_artifact_location_helper(tmp_path):
    expected_envs = [
        {"name": "WORKFLOW_ARTIFACT_PREFIX", "value": "test-op", "value_from": None},
        {
            "name": "WORKFLOW_ARTIFACT_LOCATION",
            "value": "minio://mlpipeline/artifacts/{{workflow.name}}/{{pod.name}}",
            "value_from": None,
        },
    ]

    helper = kfx.dsl._artifact_location.ArtifactLocationHelper(
        scheme="minio", bucket="mlpipeline", key_prefix="artifacts/"
    )

    @kfp.components.func_to_container_op
    def test_op(
        mlpipeline_ui_metadata: OutputTextFile(str),
        markdown_data_file: OutputTextFile(str),
    ):

        import kfx.dsl
        import kfx.vis as kfxvis

        markdown_data_file.write("### hello world")
        ui_metadata = kfxvis.kfp_ui_metadata(
            [kfxvis.markdown(kfx.dsl.kfp_artifact("markdown_data_file"))]
        )
        mlpipeline_ui_metadata.write(kfxvis.asjson(ui_metadata))
        print(ui_metadata.outputs[0].source)

    @kfp.dsl.pipeline()
    def test_pipeline():
        op: kfp.dsl.ContainerOp = test_op()
        op.apply(helper.set_envs())

        assert is_envs_similar(op.container.env, expected_envs)

    outfile = tmp_path / "pipeline.yaml"
    Compiler().compile(test_pipeline, str(outfile))
    # print(outfile.read_text())
    # assert False


def test_kfp_artifact():

    os.environ[
        kfx.dsl._artifact_location.ArtifactLocationHelper.artifact_loc_env
    ] = "gcs://your_bucket/pipelines/artifact"
    os.environ[
        kfx.dsl._artifact_location.ArtifactLocationHelper.artifact_prefix_env
    ] = "test-task"

    assert (
        kfx.dsl.kfp_artifact("some_artifact_file")
        == "gcs://your_bucket/pipelines/artifact/test-task-some-artifact.tgz"
    )

    assert (
        kfx.dsl.kfp_artifact("some_artifact_path")
        == "gcs://your_bucket/pipelines/artifact/test-task-some-artifact.tgz"
    )



"""Utils."""
import os
import os.path

from typing import Callable, NamedTuple

import kfp.dsl

from kubernetes import client as k8s_client

from kfx.dsl._compat import sanitize_k8s_name

DEFAULT_KEY_FORMAT = "{{workflow.name}}/{{pod.name}}"


class WorkflowVars(NamedTuple):
    """Describes a templated workflow environment variable."""

    name: str
    template: str


def set_workflow_env(
    workflow_vars: WorkflowVars = WorkflowVars(
        name="WORKFLOW_DEFAULT_KEY_FORMAT", template="{{workflow.name}}/{{pod.name}}"
    )
) -> Callable[[kfp.dsl.ContainerOp], kfp.dsl.ContainerOp]:
    """Modifier for kubeflow pipelines tasks.

    Setup a kfp op to pass in the workflow variables as an environment var.

    See https://github.com/argoproj/argo/blob/master/docs/variables.md

    Example::

        import kfp
        import kfx.lib.utils as kfxutils

        from kfp.components import func_to_container_op

        @func_to_container_op
        def echo_workflow_vars():
            import os

            print(os.environ.get("WORKFLOW_NAME"))


        @kfp.dsl.pipeline()
        def simple_pipeline():
            op = echo_workflow_vars()

            # op will have an environment variable "WORKFLOW_NAME"
            # that provides the name of the workflow
            op.apply(kfxutils.set_workflow_env(
                kfxutils.WorkflowVars("WORKFLOW_NAME", template="{{workflow.name}}")
            ))

    Args:
        name (str, optional): Env var name to pass in.
            Defaults to "WORKFLOW_DEFAULT_KEY_FORMAT".
        value (str, optional): Template string with argo workflow variables.
            Defaults to "{{workflow.name}}/{{pod.name}}".

    Returns:
        Callable[[kfp.dsl.ContainerOp], kfp.dsl.ContainerOp]: kfp op.
    """

    def apply_workflow_name_env(task: kfp.dsl.ContainerOp) -> kfp.dsl.ContainerOp:
        task.container.add_env_variable(
            k8s_client.V1EnvVar(name=workflow_vars.name, value=workflow_vars.template)
        )
        return task

    return apply_workflow_name_env


def set_pod_metadata_envs(
    pod_name: str = "POD_NAME",
    namespace: str = "NAMESPACE",
    node_name: str = "NODE_NAME",
) -> Callable[[kfp.dsl.ContainerOp], kfp.dsl.ContainerOp]:
    """Modifier for kubeflow pipelines tasks.

    Setup a kfp op to pass in the pod name, namespace, and node name as env
    var.

    Example::

        import kfp
        import kfx.lib.utils as kfxutils

        from kfp.components import func_to_container_op

        @func_to_container_op
        def echo_podname():
            import os

            print(os.environ.get("POD_NAME"))


        @kfp.dsl.pipeline()
        def simple_pipeline():
            op = echo_podname()
            op.apply(kfxutils.set_pod_metadata_envs())

    Args:
        pod_name (str, optional): Env var name for the pod name.
            Defaults to "POD_NAME".
        namespace (str, optional): Env var name for the namespace.
            Defaults to "NAMESPACE".
        node_name (str, optional): Env var name for the node name.
            Defaults to "NODE_NAME".

    Returns:
        Callable[[kfp.dsl.ContainerOp], kfp.dsl.ContainerOp]: kfp op.
    """

    def apply_pod_metadata_envs(task: kfp.dsl.ContainerOp) -> kfp.dsl.ContainerOp:
        for name, field_path in [
            (pod_name, "metadata.name"),
            (namespace, "metadata.namespace"),
            (node_name, "spec.nodeName"),
        ]:

            task.container.add_env_variable(
                k8s_client.V1EnvVar(
                    name=name,
                    value_from=k8s_client.V1EnvVarSource(
                        field_ref=k8s_client.V1ObjectFieldSelector(
                            field_path=field_path
                        )
                    ),
                )
            )
        return task

    return apply_pod_metadata_envs


class ArtifactLocationHelper:
    """Helper class to generate artifact location based on provided argo config.

    See an example of an `Argo configmap <https://github.com/argoproj/argo/blob/master/docs/workflow-controller-configmap.yaml>`_.

    ::

        import kfp.components
        import kfp.dsl
        import kfx.dsl


        helper = kfx.dsl.ArtifactLocationHelper(
            scheme="minio", bucket="mlpipeline", key_prefix="artifacts/"
        )

        @kfp.components.func_to_container_op
        def test_op(
            mlpipeline_ui_metadata: OutputTextFile(str), markdown_data_file: OutputTextFile(str)
        ):
            "A test kubeflow pipeline task."

            import kfx.dsl
            import kfx.vis

            # write the markdown to the `markdown-data` artifact
            markdown_data_file.write("### hello world")

            # creates an ui metadata object
            ui_metadata = kfx.vis.kfp_ui_metadata(
                # Describes the vis to generate in the kubeflow pipeline UI.
                # In this case, a markdown vis from a markdown artifact.
                [kfx.vis.markdown(kfx.dsl.kfp_artifact("markdown_data_file"))]
                # `kfp_artifact` provides the reference to data artifact created
                # inside this task
            )

            # writes the ui metadata object as the `mlpipeline-ui-metadata` artifact
            mlpipeline_ui_metadata.write(kfx.vis.asjson(ui_metadata))

            # prints the uri to the markdown artifact
            print(ui_metadata.outputs[0].source)


        @kfp.dsl.pipeline()
        def test_pipeline():
            "A test kubeflow pipeline"

            op: kfp.dsl.ContainerOp = test_op()

            # modify kfp operator with artifact location metadata through env vars
            op.apply(helper.set_envs())
    """

    artifact_loc_env: str = "WORKFLOW_ARTIFACT_LOCATION"
    artifact_prefix_env: str = "WORKFLOW_ARTIFACT_PREFIX"

    def __init__(
        self, scheme: str, bucket: str, key_prefix: str = "", key_format: str = ""
    ):
        """Creates a new instance of ArtifactLocationHelper object.

        Args:
            scheme (str): Storage scheme, e.g. s3, minio, gcs.
            bucket (str): Name of the bucket.
            key_prefix (str, optional): Key prefix set inside argo artifactory
                configmap. Will not be used if key_format is provided.
                Defaults to "".
            key_format (str, optional): Key format set inside argo artifactory
                configmap. Defaults to "".
        """

        self.scheme = scheme
        self.bucket = bucket
        self.key_prefix = key_prefix
        self.key_format = key_format

    # ref
    # https://github.com/argoproj/argo/blob/f25a45deb4a7179044034da890884432e750d98a/workflow/controller/workflowpod.go#L859
    # artifact location is defaulted using the following formula:
    #   <worflow_name>/<pod_name>/<artifact_name>.tgz
    #   (e.g. myworkflowartifacts/argo-wf-fhljp/argo-wf-fhljp-123291312382/src.tgz)
    def get_artifact_location(self) -> str:
        """Returns the path to the artifact location.

        Returns:
            str: path to the artifact location.
        """
        if self.key_format:
            return os.path.join(self.bucket, self.key_format)

        return os.path.join(self.bucket, self.key_prefix, DEFAULT_KEY_FORMAT)

    def set_envs(
        self, image: str = "e2fyi/kfx:latest"
    ) -> Callable[[kfp.dsl.ContainerOp], kfp.dsl.ContainerOp]:
        """A kfp task modifier.

        This task modifier appends 2 env variables to the task, which
        can be subsequently used to determine the artifact location and prefix.

        Args:
            image (str, optional): image to use for the task. Defaults to
                "e2fyi/kfx:latest".

        Returns:
            Callable[[kfp.dsl.ContainerOp], kfp.dsl.ContainerOp]: modified task.
        """

        def mod1(task: kfp.dsl.ContainerOp):
            task.container.image = image
            artifact_prefix = sanitize_k8s_name(task.name)
            return set_workflow_env(
                WorkflowVars(name=self.artifact_prefix_env, template=artifact_prefix)
            )(task)

        mod2 = set_workflow_env(
            WorkflowVars(
                name=self.artifact_loc_env,
                template="%s://%s" % (self.scheme, self.get_artifact_location()),
            )
        )
        return lambda task: mod2(mod1(task))


def _sanitize_artifact_name(name: str) -> str:
    """Sanitize the artifact name based on k8s resource naming convention.

    Also remove suffixes "_path" and "_file". (See this `comment <https://github.com/kubeflow/pipelines/blob/4cb81ea047361ddce7ce8b0b68133b0a92724588/sdk/python/kfp/components/_python_op.py#L327>'_.)


    Args:
        name (str): [description]

    Returns:
        str: [description]
    """
    if name.endswith("_path"):
        name = name[0 : -len("_path")]
    elif name.endswith("_file"):
        name = name[0 : -len("_file")]
    return sanitize_k8s_name(name)  # type: ignore


def kfp_artifact(name: str, ext: str = ".tgz", sanitize_name: bool = True) -> str:
    """Reference to a kfp artifact that is created within the kubeflow pipeline task.

    This function should be used inside the kfp task. It returns the artifact uri,
    which then can be provided to the kubeflow pipeline UI - i.e. as the
    `source` field inside kubeflow pipeline ui metadata.

    ::

        import kfp.components
        import kfp.dsl
        import kfx.dsl


        helper = kfx.dsl.ArtifactLocationHelper(
            scheme="minio", bucket="mlpipeline", key_prefix="artifacts/"
        )

        @kfp.components.func_to_container_op
        def test_op(
            mlpipeline_ui_metadata: OutputTextFile(str), markdown_data_file: OutputTextFile(str)
        ):
            "A test kubeflow pipeline task."

            import kfx.dsl
            import kfx.vis

            # write the markdown to the `markdown-data` artifact
            markdown_data_file.write("### hello world")

            # creates an ui metadata object
            ui_metadata = kfx.vis.kfp_ui_metadata(
                # Describes the vis to generate in the kubeflow pipeline UI.
                # In this case, a markdown vis from a markdown artifact.
                [kfx.vis.markdown(kfx.dsl.kfp_artifact("markdown_data_file"))]
                # `kfp_artifact` provides the reference to data artifact created
                # inside this task
            )

            # writes the ui metadata object as the `mlpipeline-ui-metadata` artifact
            mlpipeline_ui_metadata.write(kfx.vis.asjson(ui_metadata))

            # prints the uri to the markdown artifact
            print(ui_metadata.outputs[0].source)


        @kfp.dsl.pipeline()
        def test_pipeline():
            "A test kubeflow pipeline"

            op: kfp.dsl.ContainerOp = test_op()

            # modify kfp operator with artifact location metadata through env vars
            op.apply(helper.set_envs())


    Args:
        name (str): name of the artifact.
        ext (str, optional): extension for the artifact. Defaults to ".tgz".
        sanitize_name (bool, optional): whether to sanitize the artifact name. Defaults to True.

    Returns:
        str: uri to the artifact which can be provided to kfp ui.
    """
    artifact_loc = os.environ[ArtifactLocationHelper.artifact_loc_env]
    artifact_prefix = os.environ[ArtifactLocationHelper.artifact_prefix_env]
    artifact_name = _sanitize_artifact_name(name) if sanitize_name else name

    return os.path.join(artifact_loc, "%s-%s%s" % (artifact_prefix, artifact_name, ext))



"""Extension to kfp dsl.

::

    import kfp.components
    import kfp.dsl
    import kfx.dsl


    helper = kfx.dsl.ArtifactLocationHelper(
        scheme="minio", bucket="mlpipeline", key_prefix="artifacts/"
    )

    @kfp.components.func_to_container_op
    def test_op(
        mlpipeline_ui_metadata: OutputTextFile(str), markdown_data_file: OutputTextFile(str)
    ):
        "A test kubeflow pipeline task."

        import kfx.dsl
        import kfx.vis

        # write the markdown to the `markdown-data` artifact
        markdown_data_file.write("### hello world")

        # creates an ui metadata object
        ui_metadata = kfx.vis.kfp_ui_metadata(
            # Describes the vis to generate in the kubeflow pipeline UI.
            # In this case, a markdown vis from a markdown artifact.
            [kfx.vis.markdown(kfx.dsl.kfp_artifact("markdown_data_file"))]
            # `kfp_artifact` provides the reference to data artifact created
            # inside this task
        )

        # writes the ui metadata object as the `mlpipeline-ui-metadata` artifact
        mlpipeline_ui_metadata.write(kfx.vis.asjson(ui_metadata))

        # prints the uri to the markdown artifact
        print(ui_metadata.outputs[0].source)


    @kfp.dsl.pipeline()
    def test_pipeline():
        "A test kubeflow pipeline"

        op: kfp.dsl.ContainerOp = test_op()

        # modify kfp operator with artifact location metadata through env vars
        op.apply(helper.set_envs())

"""
from kfx.dsl._artifact_location import (
    WorkflowVars,
    ArtifactLocationHelper,
    kfp_artifact,
    set_workflow_env,
    set_pod_metadata_envs,
)



#!/usr/bin/env python
# noqa
# pylint: skip-file
"""The setup script."""

from setuptools import setup

with open("requirements.txt", "r") as filein:
    requirements = filein.readlines()

with open("requirements-dev.txt", "r") as filein:
    requirements_dev = filein.readlines()

with open("version.txt", "r") as filein:
    version = filein.read()

with open("README.md", "r") as filein:
    readme = filein.read()

setup_requirements: list = [
    "setuptools >= 41.0.0",
    "wheel >= 0.26",
]

setup(
    author="eterna2",
    author_email="eterna2@hotmail.com",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    description="Extensions for kubeflow pipeline sdk.",
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://github.com/e2fyi/kfx",
    include_package_data=True,
    package_data={"": ["version.txt", "requirements.txt", "requirements-dev.txt"]},
    keywords="kfx,kubeflow,pipelines,contrib,sdk",
    name="kfx",
    packages=["kfx.dsl", "kfx.vis"],
    setup_requires=setup_requirements,
    python_requires=">=3.6",
    install_requires=requirements,
    tests_require=requirements_dev,
    version=version,
)
